#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# V0.1 du 24/06/2019



####
#    pyFC est le diminutif de pyFileCrypter (qui était le nom originel de ce projet). Trouvant 'pyFileCrypter' un 
#    peu long, j'ai préféré le raccourcir un peu pour gagner en concision...
#
#    NB : Ce script est prévu pour fonctionner sous Linux essentiellement (mais si vous êtes motivé(s), il devrait
#    être adaptable assez facilement pour d'autres OS...)
#
#    Ce script à pour but de chiffrer/déchiffrer des fichiers ou répertoires en utilisant l'algorithme
#    du masque jetable, ou chiffre de Vernam.
#    
#    INITIALISATION:
#    Au 1er lancement du script, il faudra lui passer l'argument '-init' (puis valider cette opération par 'OK'),
#    afin que le jeu de clés de chiffrement soit créé.
#    Ce jeu consiste en 5 fichiers respectivement de 100, 200, 300, 400 et 500 Mega-Octets, contenant
#    chacun un ensemble d'octets générés aléatoirement. Les grandes tailles de ces fichiers limitent les risques
#    de copie, mais surtout d'upload par un "processus pirate" (un upload d'~ 1.5 Giga Octets prendrait beaucoup
#    de temps !). Chaque fichier clé sera généré dans une nouvelle arborescence de répertoires de 28 'étages'
#    Une fois ces fichiers clés générés, le script va vous demander d'entrer le répertoire dans lequel vous
#    souhaitez stoker ces clés.
#    La console vous indique la marche à suivre.
#    Après validation (entrée de 'OK' dans la console), le fichier pyFC.conf sera créé dans votre 
#    /home, placé dans un nouveau répertoire caché /.pyFC, puis renommé .pyFC.conf de manière
#    à devenir un fichier caché lui aussi.
#    Les droits du répertoire /.pyFC et du fichier .pyFC.conf vont être mis à 0o000 afin d'
#    améliorer un peu la sécurité (ils deviennent donc illisible tant que les droits de lecture n'auront pas 
#    été réactivés.)
#    Ces droits de lecture seront réactivés temporairement au prochain lancement du script pyFC.py (par
#    le script lui-même afin que celui-ci puisse accéder aux clés), puis re-désactivés.
#    Ce système de protection est aussi appliqué à toute l'arborescence des clés.
#
#    Pour le cas où vous souhaitez pouvoir chiffrer/déchiffrer vos fichiers sur une autre machine avec le même
#    jeu de clés, il suffira de recopier l'arborescence des clés ainsi que le fichier de configuration sur
#    ladite machine aux endroits choisis.
#    Il ne faudra en aucun cas relancer le script avec l'argument '-init', sinon celui-ci écraserait la 
#    configuration du le jeu de clés actuel avec une nouvelle configuration (les clés elles-même ne seront pas
#    effacées)! Dans une version future, si le besoin s'en fait sentir, j'implémenterai probablement la
#    possibilité d'avoir plusieurs jeux de clés (ce qui permettrait par exemple d'avoir des jeux de clés
#    partagés avec différents correspondants...)
#
#    FONCTIONNEMENT & UTILISATION :
#    Pour chaque fichier à chiffrer, le script va générer aléatoirement 5 offsets différents (1 pour chaque fichier
#    clé), puis dans chaque fichier clé aller chercher à partir de cet offset un nombre d'octets équivalent à la
#    taille du fichier à chiffrer. Ensuite, le chiffrement va s'effectuer par un XOR octet par octet comme suit :
#    temp1 = fichier_à_chiffrer ^ fichier_clé1[offset1, size]
#    temp2 = temp1 ^ fichier_clé2[offset2, size]
#    temp3 = temp2 ^ fichier_clé3[offset3, size]
#    temp4 = temp3 ^ fichier_clé4[offset4, size]
#    temp5 = temp4 ^ fichier_clé5[offset5, size]
#
#
#    Il est aussi nécessaire de stocker les offsets utilisés au sein de chaque fichier chiffré. Ces offsets
#    sont codés sur 4 octets (nombre hexa 32bits).
#    A partir de ces données, le script va créer le header du fichier une fois chiffré :
#
#    XXXX -> 4 octets representant l'offset1 : offset1 ^ fichier_clé1 [10000 +  29*(TailleFichier % 9999), 4]
#                                                      ^ fichier_clé2 [20000 + 113*(TailleFichier % 8888), 4]
#                                                        -> Poids des octets dans le codage : 0231
#    XXXX -> 4 octets representant l'offset2 : offset2 ^ fichier_clé2 [30000 + 229*(TailleFichier % 7777), 4]
#                                                      ^ fichier_clé3 [40000 + 349*(TailleFichier % 6666), 4]
#                                                        -> Poids des octets dans le codage : 0231
#    XXXX -> 4 octets representant l'offset3 : offset3 ^ fichier_clé3 [50000 + 463*(TailleFichier % 5555), 4]
#                                                      ^ fichier_clé4 [60000 + 601*(TailleFichier % 4444), 4]
#                                                        -> Poids des octets dans le codage : 0231
#    XXXX -> 4 octets representant l'offset4 : offset4 ^ fichier_clé4 [70000 + 733*(TailleFichier % 3333), 4]
#                                                      ^ fichier_clé5 [80000 + 863*(TailleFichier % 2222), 4]
#                                                        -> Poids des octets dans le codage : 0231
#    XXXX -> 4 octets representant l'offset5 : offset5 ^ fichier_clé5 [90000 + 1013*(TailleFichier % 1111), 4]
#                                                      ^ fichier_clé1 [100000 + 1151*(TailleFichier % 5432), 4]
#                                                        -> Poids des octets dans le codage : 0231
#    
#    Une fois ce header généré, il sera mis en entête de temp5 (voir plus haut), puis le fichier sera sauvegardé
#    sous le nom int([time]).EXT ([time] étant le nombre de secondes écoulées depuis le 01/01/1970
#    EXT aura pour valeur l'index '000', '001', '002' selon le rang du fichier compressé correspondant à
#    un même fichier non compressé d'origine (cas des fichiers > 100MBytes)
#    Dans le cas d'une compression en plusieurs 'volumes' (.000, .001, etc...), le time servant à définir
#    le nom de fichier devra être le même.
#
#
#    CAS DU CHIFFREMENT D'UN REPERTOIRE :
#    Plutôt qu'un simple fichier, pyFC peut chiffrer un répertoire complet en y incluant tous ses
#    fichiers et sous-répertoires. Pour cela, le répertoire à chiffrer sera dans un premier temps
#    compressé en fichier zip, puis ensuite traité comme un simple fichier normal.
#    NB : Il en est de même maintenant pour le chiffrement d'un fichier 'simple'. Celui-ci sera
#    préalablement archivé dans un zip, avant que ce zip soit chiffré. Cela présente l'avantage de
#    ne pas avoir à stocker le nom d'origine du fichier (puisqu'il figurera dans l'archive zip).
#
#
#    ARGUMENTS :
#    Pour chiffrer un fichier / répertoire : --> python3 pyFC.py -e [chemin complet vers le fichier
#                                                                         ou répertoire à chiffrer]
#    Pour déchiffrer un fichier            : --> python3 pyFC.py -d [chemin complet vers le fichier
#                                                                             chiffré à déchiffrer]
#
#
#
####




import os
import sys
import shutil
import secrets
import time
import random
import string
import zipfile





def setConfFileAccess(accessibility):
    ''' Permet de rendre /home/user/.pyFC/.pyFC.conf visible et accessible ou non selon les besoins'''
    if accessibility:
        os.chmod(pfcDir, 0o774) # On met les droits de lecture/écriture sur le répertoire
        os.chmod(pfcConfFile, 0o774) # On met les droits de lecture/écriture sur le fichier conf
        #todo : gérer le attrib sous win
    else:
        os.chmod(pfcDir, 0o774) # On met les droits de lecture/écriture sur le répertoire
        try:
            os.chmod(pfcConfFile, 0o000)# On supprime les droits de lecture/écriture sur le répertoire
            os.chmod(pfcDir, 0o000)# On supprime les droits de lecture/écriture sur le répertoire
            #todo : gérer le attrib sous win
        except PermissionError:
            pass



def setKeyfilesAccess(accessibility):
    ''' Permet de rendre les 5 fichiers clé keyFilePath visibles et accessibles ou non selon les besoins.'''
    nindex = 0
    for fullPath in keyFilepathList:
        if accessibility:  # On rend les clés accessibles
            fullPath = fullPath[0:len(fullPath)-145]
            for n in range(29):
                dirStr = keyFilepathList[nindex][0:len(fullPath)+(n+1)*5]
                os.chmod(dirStr, 0o774)
                #todo : gérer le attrib sous win
        else:  # On rend les clés inaccessibles
            for n in range(29):
                dirStr = keyFilepathList[nindex][0:len(fullPath)-n*5]
                try:
                    os.chmod(dirStr, 0o000)
                    #todo : gérer le attrib sous win
                except PermissionError:
                    pass
        nindex += 1



def removeRNFromListe(listName):
    ''' Enlève les caractères \r ou \r\n en fin de ligne pour chaque string contenue dans listName '''
    newList = []
    for oneLine in listName:
        if oneLine.endswith("\r\n"):
            oneLine = oneLine[0:(len(oneLine)-2)]
        if oneLine.endswith("\n"):
            oneLine = oneLine[0:(len(oneLine)-1)]
        if len(oneLine) >0:
            newList.append(oneLine)
    return newList



def readFileLines(filename):
    ''' Lit le fichier filename (fichier texte), et stocke chacune de ses lignes dans la <list> newList ''' 
    if os.path.exists(filename):
        oneStringList = []
        with open(filename, 'r') as myFile:
            oneStringList = myFile.readlines()
        newList = removeRNFromListe(oneStringList)
        return newList
    else:
        return ("Le fichier "+filename+" n'existe pas!")



def readConfFile():
    ''' Lit le fichier /home/user/.pyFC/.pyFC.conf et 
        stocke son contenu ligne par ligne dans keyFilepathList '''
    global keyFilepathList
    setConfFileAccess(True)
    keyFilepathList = readFileLines(pfcConfFile)
    setConfFileAccess(False)



def print4(message):
    ''' Simple fonction print() affichant une ligne avant le message et deux lignes après '''
    print()
    print(message)
    print()
    print()



def createKeyFiles():
    ''' Permet de créer les 5 fichiers clés (fichiers 'binaires') contenant nbBytes générés aléatoirement. '''
    global pfcConfFile
    global pfcDir
    global keyFilepathList
    print("Création d'un nouveau jeu de clés de chiffrement. Cela peut prendre un moment...")
    pathToScript = os.path.dirname(os.path.realpath(sys.argv[0]))

    # Création de 5 arborescences contenant 28 niveaux de répertoires
    # Chaque fichier clé sera généré dans le répertoire le plus loin de la racine (niveau 28)
    for keyIndex in range(5):
        for rep in range(28):
            newrepName = '/.'
            rangeSize = 3
            if rep == 0:
                newrepName += str(keyIndex+1)
                rangeSize = 2;
            for _ in range(rangeSize):
                newrepName += random.choice(string.digits+string.ascii_letters)
            pathToScript += newrepName
            os.mkdir(pathToScript)
            os.chmod(pathToScript, 0o774)
        nbBytes = 100*(keyIndex+1)*(2**20)
        myBytes = os.urandom(nbBytes)
        oneFileName = pathToScript+'/.key'+str(keyIndex+1)
        with open(oneFileName, 'wb') as keyFile:  # On écrit les 5 fichiers 'clé'
            keyFile.write(myBytes)
        print("Fichier key"+str(keyIndex+1), "créé")
        pathToScript = os.path.dirname(os.path.realpath(sys.argv[0]))
        keyFilepathList.append(oneFileName)  # On mémorise le chemin vers chaque clé
    
    print()
    print("Les 5 nouvelles clés de chiffrement ont été créées. Pour des raisons de sécurité, il est")
    print("préférable de placer ces fichiers clés ailleurs que dans le répertoire .../pyFile")
    print("Néanmoins, le programme fonctionnera correctement si vous n'entrez rien (simple appui sur [ENTER].")
    print("Dans ce cas, les fichiers clés resteront simplement dans .../pyFile")
    print()
    print("Pour la ligne ci-dessous, si vous êtes sous Windows, veillez à")
    print("remplacer les \ par des / dans la définition du chemin.")
    print()
    
    newPathStr = input("Entrez le chemin où sera installé le nouveau jeu de clés : ")
    if len(newPathStr) > 0:
        while not os.path.exists(newPathStr):
            print(newPathStr,"n'est pas un répertoire valide !")
            newPathStr = input("Entrez le chemin où sera installé le nouveau jeu de clé : ")
            print()
        if '/' in newPathStr[-1]:  # Pour éliminer le / final si présent
            newPathStr = newPathStr[0:len(newPathStr)-1]

        # Déplacement des clés vers leur nouveau répertoire et M.A.J de keyFilepathList
        for nindex in range(5):
            end = len(keyFilepathList[nindex])-141
            print(keyFilepathList[nindex][0:end])
            print()
            print("Déplacement du fichier key"+str(nindex))
            shutil.move(keyFilepathList[nindex][0:end], newPathStr)
            keyFilepathList[nindex] = newPathStr+keyFilepathList[nindex][end-5:-1]+str(nindex+1)
        print()

    # Création de /home/user/.pyFC si nécessaire & gestion des droits...
    if not os.path.exists(pfcDir):
        os.mkdir(pfcDir)
    os.chmod(pfcDir, 0o774) # On met les droits de lecture/écriture sur le répertoire
    if os.path.exists(pfcConfFile):
        os.chmod(pfcConfFile, 0o774) # On met les droits de lecture/écriture sur le fichier conf
        os.remove(pfcConfFile)  # Puis on efface le fichier existant (obsolète)

    for nindex in range(5):
        with open(pfcConfFile, 'a') as confFile:  # On reporte keyFilepathList dans pyFC.conf
            confFile.write(keyFilepathList[nindex]+'\r\n')

    # La configuration est maintenant écrite dans /home/user/.pyFC/.pyFC.conf,
    # On va lire la configuration pour protéger les répertoires qui contiennent les clés
    readConfFile()
    # on protège les répertoires des clés en supprimant les droits de lecture de ces répertoires
    setKeyfilesAccess(False)
    print4("Le nouveau jeu de clés est maintenant activé")


def printUsage():
    ''' Affiche l'aide, le 'mode d'emploi' de pyFC '''
    print()
    print("Utilisation :")
    print("pyFC.py              --> affiche ce mode d'emploi")
    print("pyFC.py -help        --> affiche ce mode d'emploi")
    print("pyFC.py -init        --> pour initialiser pyFC avec un nouveau jeu de clé de chiffrement")
    print("pyFC.py -e filename  --> pour chiffrer (-e --> encrypt) le fichier nommé 'filename'")
    print("pyFC.py -e dirname   --> pour chiffrer (-e --> encrypt) le répertoire nommé 'dirname'")
    print("pyFC.py -d filename  --> pour déchiffrer (-d --> decrypt) le fichier nommé 'filename'. Ici, l'")
    print("                         argument filename peut désigner soit un fichier, soit un répertoire")
    print("                         contenant d'autres fichiers ou répertoires")
    print("pyFC.py -showconf    --> change les droits sur ~/.pyFC/.pyFC.conf pour le rendre accessible") 
    print("                         --> os.chmod(0o774)") 
    print("pyFC.py -hideconf    --> change les droits sur ~/.pyFC/.pyFC.conf pour le rendre inaccessible") 
    print("                         --> os.chmod(0o000)") 
    print("                     Attention, les arguments -showconf et -hideconf ne doivent être")
    print("                     utilisés qu'après la mise en place d'un jeu de clé valide !")
    print()
    print()



def XOR2Arrays(bytesArray1, bytesArray2):    
    ''' Effectue un XOR byte à byte sur les 2 tableaux de bytes transmis,
        puis retourne le bytearray qui en resulte '''
    return bytearray(map(ord, ''.join(chr(a ^ b) for a,b in zip(bytesArray1, bytesArray2))))



def setCryptoBytes(mode, fileLength, offsetIntArray):
    ''' Cette fonction permet de créer un tableau de bytes qui sera utilisé pour
        le chiffrement/déchiffrement du fichier à traiter '''
    cryptoArray = []
    with open(keyFilepathList[0], 'rb') as currentKeyFile1:
        if not 'decryptData' in mode:
            currentKeyFile1.seek(10000 + 29*(fileLength % 9999))
            byteArrayOff11 = currentKeyFile1.read(4)
            currentKeyFile1.seek(100000 + 1151*(fileLength % 5432))
            byteArrayOff52 = currentKeyFile1.read(4)
        if not 'decryptOffset' in mode:
            currentKeyFile1.seek(offsetIntArray[0])
            byteArrayData1 = currentKeyFile1.read(fileLength)
    with open(keyFilepathList[1], 'rb') as currentKeyFile2:
        if not 'decryptData' in mode:
            currentKeyFile2.seek(20000 + 113*(fileLength % 8888))
            byteArrayOff12 = currentKeyFile2.read(4)
            currentKeyFile2.seek(30000 + 229*(fileLength % 7777))
            byteArrayOff21 = currentKeyFile2.read(4)
        if not 'decryptOffset' in mode:
            currentKeyFile2.seek(offsetIntArray[1])
            byteArrayData2 = currentKeyFile2.read(fileLength)
    with open(keyFilepathList[2], 'rb') as currentKeyFile3:
        if not 'decryptData' in mode:
            currentKeyFile3.seek(40000 + 349*(fileLength % 6666))
            byteArrayOff22 = currentKeyFile3.read(4)
            currentKeyFile3.seek(50000 + 463*(fileLength % 5555))
            byteArrayOff31 = currentKeyFile3.read(4)
        if not 'decryptOffset' in mode:
            currentKeyFile3.seek(offsetIntArray[2])
            byteArrayData3 = currentKeyFile3.read(fileLength)
    with open(keyFilepathList[3], 'rb') as currentKeyFile4:
        if not 'decryptData' in mode:
            currentKeyFile4.seek(60000 + 601*(fileLength % 4444))
            byteArrayOff32 = currentKeyFile4.read(4)
            currentKeyFile4.seek(70000 + 733*(fileLength % 3333))
            byteArrayOff41 = currentKeyFile4.read(4)
        if not 'decryptOffset' in mode:
            currentKeyFile4.seek(offsetIntArray[3])
            byteArrayData4 = currentKeyFile4.read(fileLength)
    with open(keyFilepathList[4], 'rb') as currentKeyFile5:
        if not 'decryptData' in mode:
            currentKeyFile5.seek(80000 + 863*(fileLength % 2222))
            byteArrayOff42 = currentKeyFile5.read(4)
            currentKeyFile5.seek(90000 + 1013*(fileLength % 1111))
            byteArrayOff51 = currentKeyFile5.read(4)
        if not 'decryptOffset' in mode:
            currentKeyFile5.seek(offsetIntArray[4])
            byteArrayData5 = currentKeyFile5.read(fileLength)
    
    if not 'decryptData' in mode:
        cryptoArray.append(byteArrayOff11)
        cryptoArray.append(byteArrayOff12)
        cryptoArray.append(byteArrayOff21)
        cryptoArray.append(byteArrayOff22)
        cryptoArray.append(byteArrayOff31)
        cryptoArray.append(byteArrayOff32)
        cryptoArray.append(byteArrayOff41)
        cryptoArray.append(byteArrayOff42)
        cryptoArray.append(byteArrayOff51)
        cryptoArray.append(byteArrayOff52)
    if not 'decryptOffset' in mode:
        cryptoArray.append(byteArrayData1)
        cryptoArray.append(byteArrayData2)
        cryptoArray.append(byteArrayData3)
        cryptoArray.append(byteArrayData4)
        cryptoArray.append(byteArrayData5)
    return cryptoArray



def encrypt(savedName, filesAsByteArray, fileIndex, nbSlices):
    ''' Fonction qui permet de procéder au chiffrement du tableau de bytes
        'filesAsByteArray' et de le sauver sous le nom savedName '''
    setKeyfilesAccess(True)
    offsetIntArray = []
    offsetBytesArray = []

    # ici, on créé le tableau d'offsets 
    # On aura toujours len(filesAsByteArray) < keyFileSize (donc < 100 MBytes)
    for index in range(5):
        intOffset = secrets.randbelow(100*(index+1)*(2**20)-len(filesAsByteArray))    
        byte3 = intOffset >> 24
        byte2 = (intOffset - byte3*(256**3)) >> 16
        byte1 = (intOffset - byte3*(256**3) - byte2*(256**2)) >> 8
        byte0 = (intOffset - byte3*(256**3) - byte2*(256**2) - byte1*256)
        bytesList = [byte0, byte2, byte3, byte1]
        offsetIntArray.append(intOffset) # offsetIntArray contient les offsets X au format integer normal
        offsetBytesArray.append(bytearray(bytesList)) # offsetBytesArray contient les offsets X au format 0231

    # On va créer le byteArray à partir duquel on va chiffrer les offsets et les données
    cryptoArray = setCryptoBytes('encrypt', len(filesAsByteArray), offsetIntArray)

    # On procède au chiffrement des offsets
    headerXOffset = []
    headerXOffset.append(XOR2Arrays(XOR2Arrays(offsetBytesArray[0], cryptoArray[0]), cryptoArray[1]))
    headerXOffset.append(XOR2Arrays(XOR2Arrays(offsetBytesArray[1], cryptoArray[2]), cryptoArray[3]))
    headerXOffset.append(XOR2Arrays(XOR2Arrays(offsetBytesArray[2], cryptoArray[4]), cryptoArray[5]))
    headerXOffset.append(XOR2Arrays(XOR2Arrays(offsetBytesArray[3], cryptoArray[6]), cryptoArray[7]))
    headerXOffset.append(XOR2Arrays(XOR2Arrays(offsetBytesArray[4], cryptoArray[8]), cryptoArray[9]))

    # On procède au chiffrement des données du fichier
    print("Tranche "+str(fileIndex+1)+"/"+str(nbSlices)+":")
    print("--> Chiffrement clé 1/5")
    tempData1 = XOR2Arrays(filesAsByteArray, cryptoArray[10])
    print("--> Chiffrement clé 2/5")
    tempData2 = XOR2Arrays(tempData1, cryptoArray[11])
    print("--> Chiffrement clé 3/5")
    tempData3 = XOR2Arrays(tempData2, cryptoArray[12])
    print("--> Chiffrement clé 4/5")
    tempData4 = XOR2Arrays(tempData3, cryptoArray[13])
    print("--> Chiffrement clé 5/5")
    finalData = XOR2Arrays(tempData4, cryptoArray[14])

    # On sauve le fichier chiffré
    with open(savedName, 'wb') as treatedFile:
        for oneOffset in headerXOffset:
            treatedFile.write(oneOffset)
        treatedFile.write(finalData)
    print("Chiffrement terminé -> Création du fichier " + savedName )
    print()    
    setKeyfilesAccess(False)



def addDirToZip(zipHandle, path, basePath=""):
    ''' On va ici placer dans l'archive zipHandle tout le contenu du répertoire path '''
    basePath = basePath.rstrip("\\/") + ""
    basePath = basePath.rstrip("\\/")
    for root, dirs, files in os.walk(path):
        # add dir itself (needed for empty dirs)
        zipHandle.write(os.path.join(root, "."))
        # add files
        for oneFile in files:
            filePath = os.path.join(root, oneFile)
            inZipPath = filePath.replace(basePath, "", 1).lstrip("\\/")
            #print filePath + " , " + inZipPath
            zipHandle.write(filePath, inZipPath)



def sliceAndEncrypt(filePath):
    ''' Pour simplifier les choses au niveau de la fonction de chiffrement (encrypt(filePath)), on impose
        que pour tout fichier à chiffrer, la taille  doit être inférieure à celle de la clé de chiffement
        la plus petite, soit 100MBytes (100 * 2**20). Si le fichier à chiffrer est de taille supérieure, 
        on va alors ici le découper en plusieurs tranches de tailles ~ équivalentes et inférieures à cette
        limite de 100 MBytes. On chiffrera ensuite l'ensemble des fichiers obtenus
        La variable limitSize (voir un peu plus bas après la ligne 'if proceed:') contiendra la limite de
        taille réellement choisie '''
    
    print()
    proceed = False
    # Si filePath pointe vers un fichier, on va le zipper à partir du fichier avant le chiffrement
    # cela permettra de ne pas avoir à se préoccuper du nom du fichier original (qui sera contenu dans le zip)
    if os.path.isfile(filePath):  
        # On va créer le zip dans le répertoire courant (currentDir)
        # Attention à bien avoir les droits d'écriture dans ce répertoire !
        simpleFilename = filePath.split('/')[-1] # Dans l'archive, on ne mettra que le nom du fichier, sans son path
        print("Chiffrement de " +simpleFilename)
        print()
        print("Étape 1 : Création de l'archive principale")
        zipHandle = zipfile.ZipFile('crypted.zip', 'w', zipfile.ZIP_DEFLATED)
        zipHandle.write(simpleFilename) 
        zipHandle.close()
        proceed = True
   
    # Si filePath pointe vers un répertoire, on va le zipper afin de pouvoir le traiter comme un fichier
    if os.path.isdir(filePath):
        # On va créer le zip dans le répertoire courant (currentDir)
        # Attention à bien avoir les droits d'écriture dans ce répertoire !
        zipHandle = zipfile.ZipFile('crypted.zip', 'w', zipfile.ZIP_DEFLATED)
        savedCurrentDir = currentDir
        repList = filePath.split('/')
        fRepList = []
        repToGo = ''
        proceed = True

        # Si l'on veut par exemple crypter le répertoire /home/toto/Documents , on va se placer dans /home/toto de
        # manière à se trouver à un étage 'supérieur' au répertoire 'Documents'. Ainsi, on pourra plus facilement
        # inclure ce nom de répertoire (et uniquement lui) dans l'archive zip, qui contiendra alors le répertoire
        #'/Documents' dans lequel on retrouvera tous ses fichiers et sous-répertoires
        for nStr in repList:
            if len(nStr)>0:
                fRepList.append(nStr)
                zipThisDir = nStr # Selon notre exemple, zipThisDir contiendrait 'Documents'
       
        if len(fRepList) == 0:
            proceed = False
            print4("Erreur, vous ne pouvez pas chiffrer l'ensemble de la racine (/) ! \r\n \
                    Sélectionnez un répertoire existant !")
        
        if len(fRepList) == 1:
            repToGo = '/'
            os.chdir(repToGo) 
            print("Chiffrement de " +filePath)
            print()
            print("Étape 1 : Création de l'archive principale")
            addDirToZip(zipHandle, zipThisDir) 
            zipHandle.close()
            os.chdir(savedCurrentDir) 
        
        if len(fRepList)>1:
            del fRepList[-1]
            for oneDir in fRepList:
                repToGo += '/' + oneDir
            os.chdir(repToGo) # Selon notre exemple, on remonte ici dans le répertoire '/home/toto'
            print("Chiffrement de " +filePath)
            print()
            print("Étape 1 : Création de l'archive principale")
            addDirToZip(zipHandle, zipThisDir) # Ici, on va ajouter à l'archive zip tout le
            zipHandle.close()                  # contenu de 'Documents' vu de '/home/toto'
            os.chdir(savedCurrentDir) # On retourne ici dans le répertoire d'où l'on a appelé le script

    if proceed :
        print()
        print("Étape 2 : Découpe et Chiffrement :")
        filePath = currentDir+'/crypted.zip'
        fileSize = os.path.getsize(filePath)

        # On va charger le fichier à traiter, puis le découper en byteArray(s) de taille < 100MBytes si besoin
        # On a le choix ici de la limite de taille des byteArrays (doit être impérativement < 100MBytes)
        # Plus on abaisse cette limite, plus le chiffrement sera rapide pour les fichiers de taille supérieure
        # En contrepartie, on aura un chiffrement en plusieurs étapes (autant que de tranches, après découpe
        # du fichier crypted.zip)
        limitSize = (3+int(fileSize/(15*(2**20)))) * (2**20)
        if limitSize > (50*(2**20)):
            limitSize = 50*(2**20)
        print(limitSize)
        if fileSize < limitSize:
            nbSlices = 1
            rawSliceSize = fileSize
        else:
            nbSlices = int(fileSize/limitSize) + 1
            rawSliceSize = int(fileSize/nbSlices) + 1
      
        filesAsByteArray = []
        with open(filePath, 'rb') as treatedFile:
            for n in range(nbSlices):
                treatedFile.seek(n*rawSliceSize)
                filesAsByteArray.append(treatedFile.read(rawSliceSize))

        # On va maintenant chiffrer chaque byteArray, puis le sauver en tant que fichier chiffré
        namePart = '/' + str(int(time.time()))
        print(nbSlices)
        for n in range(nbSlices):
            fileExt = ".00"
            if 9 < n < 100:
                fileExt = ".0"
            if 1000 > n > 99:
                fileExt = "."        
            saveName = currentDir + namePart + fileExt + str(n)
            encrypt(saveName, filesAsByteArray[n], n, nbSlices)
            
        # On va supprimer le fichier crypted.zip qui n'est plus utile
        os.remove(filePath)
        time.sleep(2) # Petite pause avant fermeture du terminal si lancé en clic-droit depuis caja



def decrypt(filePath, nbSlices):
    ''' Fonction qui permet de procéder au déchiffrement du fichier filePath '''
    if os.path.isfile(filePath):  # Si filePath pointe bien vers un fichier
        setKeyfilesAccess(True)
        usedLength = os.path.getsize(filePath) - 20 # On retire les 20 octets d'entête qui ont été 
        headerXOffset = []                          # ajouté lors du chiffrement

        # On va reconstituer le byteArray à partir duquel on a chiffré les offsets
        offsetIntArray = 0
        cryptoArray = setCryptoBytes('decryptOffset', usedLength, offsetIntArray)

        # Lecture et déchiffrement 1/2 des 5 offsets chiffrés
        with open(filePath, 'rb') as cryptedFile:
            off1 = cryptedFile.read(4)
            headerXOffset.append(XOR2Arrays(XOR2Arrays(off1, cryptoArray[0]), cryptoArray[1]))
            cryptedFile.seek(4)
            off2 = cryptedFile.read(4)
            headerXOffset.append(XOR2Arrays(XOR2Arrays(off2, cryptoArray[2]), cryptoArray[3]))
            cryptedFile.seek(8)
            off3 = cryptedFile.read(4)
            headerXOffset.append(XOR2Arrays(XOR2Arrays(off3, cryptoArray[4]), cryptoArray[5]))
            cryptedFile.seek(12)
            off4 = cryptedFile.read(4)
            headerXOffset.append(XOR2Arrays(XOR2Arrays(off4, cryptoArray[6]), cryptoArray[7]))
            cryptedFile.seek(16)
            off5 = cryptedFile.read(4)
            headerXOffset.append(XOR2Arrays(XOR2Arrays(off5, cryptoArray[8]), cryptoArray[9]))
            cryptedFile.seek(20)
            cypherData = cryptedFile.read(usedLength)

        # Reconstitution des offsets (Déchiffrement 2/2)
        intOffset = []
        for oneHeaderOffset in headerXOffset:
            byte0 = oneHeaderOffset[0]
            byte2 = oneHeaderOffset[1]
            byte3 = oneHeaderOffset[2]
            byte1 = oneHeaderOffset[3]
            intOffset.append(int(byte3*(256**3) + byte2*(256**2)+ byte1*256 + byte0))

        # On va reconstituer le byteArray à partir duquel on a chiffré les données
        cryptoArray = setCryptoBytes('decryptData', usedLength, intOffset)

        # On procède au déchiffrement des données
        fileExt = filePath.split('.')[-1]
        intFileExt = int(fileExt)
        print("Tranche "+str(intFileExt+1)+"/"+str(nbSlices)+":")
        print("--> Déchiffrement clé 1/5")
        tempData1 = XOR2Arrays(cypherData, cryptoArray[4])
        print("--> Déchiffrement clé 2/5")
        tempData2 = XOR2Arrays(tempData1, cryptoArray[3])
        print("--> Déchiffrement clé 3/5")
        tempData3 = XOR2Arrays(tempData2, cryptoArray[2])
        print("--> Déchiffrement clé 4/5")
        tempData4 = XOR2Arrays(tempData3, cryptoArray[1])
        print("--> Déchiffrement clé 5/5")
        finalData = XOR2Arrays(tempData4, cryptoArray[0])

        # On sauve le fichier déchiffré
        with open(currentDir+'/decypher.'+fileExt, 'wb') as treatedFile:
            treatedFile.write(finalData)
        print("Déchiffrement terminé -> Écriture du fichier " + currentDir+'/decypher.'+fileExt)
        print()    
        setKeyfilesAccess(False)
        return(currentDir+'/decypher.'+fileExt) # On va retourner le nom du fichier déchiffré, 
                                                # afin de pouvoir le ré-assembler si besoin



def getSlicesAndDecrypt(filePath):
    ''' On va déterminer ici en combien de tranches le fichier original a été découpé avant chiffrement.
        On devra déchiffrer chaque tranche, puis ré-assembler le tout pour reconstituer l'archive originale.
        Dans un dernier temps, on pourra enfin décompresser l'archive '''

    print()
    if os.path.exists(filePath):
        print("Déchiffrement de " +filePath)
        print()
        repList = filePath.split('.')
        if '000' in repList[-1]: # On s'assure que le fichier à déchiffrer possède bien une extension en .000
            myCypherList = []
            myCypherList.append(filePath)
            index = 1
            bitContinue = True
            while bitContinue:
                fileExt = ".00"
                if 9 < index < 100:
                    fileExt = ".0"
                if 1000 > index > 99:
                    fileExt = "."        
                tryThisFile = repList[0]+fileExt+str(index)
                index += 1
                if os.path.exists(tryThisFile):
                    myCypherList.append(tryThisFile) # myCypherList contiendra la liste des fichiers à déchiffrer
                else:
                    bitContinue = False
            decypheredList = []
            decypheredByteArray = []
            for oneCryptedSlice in myCypherList:
                decypheredList.append(decrypt(oneCryptedSlice, len(myCypherList)))
            for oneFile in decypheredList:
                with open(oneFile, 'rb') as decypheredSlice:
                    tmpArray = decypheredSlice.read()
                    decypheredByteArray.append(tmpArray)
                os.remove(oneFile)
            with open('finalFile.zip', 'wb') as finalFile:
                for oneByteArray in decypheredByteArray:
                    finalFile.write(oneByteArray)
            finalZip = zipfile.ZipFile('finalFile.zip', 'r')
            try:
               rootArcName = finalZip.namelist()[0] # On vient chercher le nom du répertoire racine contenu
            except:                                 # dans finalFile.zip
               print("L'archive semble vide...")
               print()
            finalZip.extractall(currentDir)
            os.remove('finalFile.zip')
            print4("Extraction / création de " +rootArcName+ "  --> Déchiffrement terminé")
            time.sleep(2) # Petite pause avant fermeture du terminal si lancé en clic-droit depuis caja
        else:
            print4(filePath+" n'est pas un fichier valide pour déchiffrement !")
            time.sleep(2) # Petite pause avant fermeture du terminal si lancé en clic-droit depuis caja
        



###################### Lancement du programme principal ######################
#
if __name__ == "__main__":
    pfcDir = os.path.expanduser('~')+'/.pyFC'
    pfcConfFile = pfcDir+'/.pyFC.conf'
    currentDir = os.getcwd()
    keyFilepathList = []
    if len(sys.argv) == 1:
        printUsage()
    if len(sys.argv) == 2:
        foundBit2 = False
        if '-help' in (sys.argv[1]):
            foundBit2 = True
            printUsage()
        if '-init' in (sys.argv[1]):
            foundBit2 = True
            print("Cette opération va générer un nouveau jeu de clés de chiffrement.")
            strOK = input("Voulez-vous vraiment continuer ? (Abandon si différent de OK) : ")
            if len(strOK)==2 and 'OK' in strOK:
                createKeyFiles()
            else:
                print4("Abandon...")
        if '-showconf' in (sys.argv[1]):
            foundBit2 = True
            readConfFile()
            setConfFileAccess(True)
            setKeyfilesAccess(True)
            print4("Configuration et arborescence de clés accessibles.")
        if '-hideconf' in (sys.argv[1]):
            foundBit2 = True
            readConfFile()
            setConfFileAccess(False)
            setKeyfilesAccess(False)
            print4("Configuration et arborescence de clés protégées.")
        if not foundBit2: # Cas où l'on oublie par exemple de fournir un chemin pour le chiffrement/déchiffrement
            printUsage()
    if len(sys.argv) == 3:
        ed = False
        badFilePath = False
        if '-e' in (sys.argv[1]) or '-d' in (sys.argv[1]):
            ed = True
            if len(sys.argv[2]) > 1:
                if os.path.exists(sys.argv[2]):  # On va ici traiter le chiffrement ou le déchiffrement 
                    readConfFile()  # On lit le fichier conf pour accéder aux chemins des fichiers clés
                    if '-e' in (sys.argv[1]):
                        sliceAndEncrypt(sys.argv[2])
                    if '-d' in (sys.argv[1]):
                        getSlicesAndDecrypt(sys.argv[2])
                else:
                    badFilePath = True
        if not ed:
            # Erreur sur argument -e ou -d
            print4("Erreur : commande incompréhensible !")
            printUsage()
        if badFilePath:
            # Erreur sur chemin vers fichier ou répertoire passé en argument
            print4("Erreur : la chaine à traiter n'existe pas, n'est ni un fichier ni un répertoire !")
    if len(sys.argv) > 3: # Trop de paramètres passés en argument !
        print4("Erreur : trop de paramètres passés en argument !")
        printUsage()

